import builtins
import typing

class FieldIndex:
    """
    *Internal*

    Instances of the `FieldIndex` class represents a field index created by [`vector_index`](#vector-index), [`keyword_index`](#keyword-index), or [`semantic_index`](#semantic-index) functions.
    """

    ...

class FieldSpec:
    """
    *Internal*

    Instances of the `FieldSpec` class represents a field specification created by [`text`](#text), [`int`](#int), [`float`](#float), [`bool`](#bool), [`f32_vector`](#f32-vector), [`u8_vector`](#u8-vector), [`i8_vector`](#i8-vector), [`binary_vector`](#binary-vector), [`f32_sparse_vector`](#f32-sparse-vector), [`u8_sparse_vector`](#u8-sparse-vector), [`bytes`](#bytes), or [`list`](#list) functions.
    """

    def required(self) -> FieldSpec:
        """
        Mark a field as required. All fields are optional by default.

        Example:

        ```python
        from topk_sdk.schema import text

        client.collections().create("books", schema={
            "title": text().required()
        })
        ```
        """
        ...
    def optional(self) -> FieldSpec: ...
    def index(self, index: FieldIndex) -> FieldSpec:
        """
        Create an index on a field.

        Example:

        ```python
        from topk_sdk.schema import text, keyword_index

        client.collections().create("books", schema={
            "title": text().index(keyword_index())
        })
        ```
        """
        ...

# data types
def text() -> FieldSpec:
    """
    Create a [FieldSpec](https://docs.topk.io/sdk/topk-py/schema#FieldSpec) type for `text` values.

    Example:

    ```python
    from topk_sdk.schema import text

    client.collections().create("books", schema={
        "title": text()
    })
    ```
    """
    ...

def int() -> FieldSpec:
    """
    Create an integer field specification.
    """
    ...

def float() -> FieldSpec:
    """
    Create a [FieldSpec](https://docs.topk.io/sdk/topk-py/schema#FieldSpec) type for `float` values.

    Example:

    ```python
    from topk_sdk.schema import float

    client.collections().create("books", schema={
        "price": float()
    })
    ```
    """
    ...

def bool() -> FieldSpec:
    """
    Create a [FieldSpec](https://docs.topk.io/sdk/topk-py/schema#FieldSpec) type for `bool` values.

    Example:

    ```python
    from topk_sdk.schema import bool

    client.collections().create("books", schema={
        "is_published": bool()
    })
    ```
    """
    ...

def f32_vector(dimension: builtins.int) -> FieldSpec:
    """
    Create a [FieldSpec](https://docs.topk.io/sdk/topk-py/schema#FieldSpec) type for `f32_vector` values.

    Example:

    ```python
    from topk_sdk.schema import f32_vector

    client.collections().create("books", schema={
        "title_embedding": f32_vector(dimension=1536)
    })
    ```
    """
    ...

def u8_vector(dimension: builtins.int) -> FieldSpec:
    """
    Create a [FieldSpec](https://docs.topk.io/sdk/topk-py/schema#FieldSpec) type for `u8_vector` values.

    Example:

    ```python
    from topk_sdk.schema import u8_vector

    client.collections().create("books", schema={
        "title_embedding": u8_vector(dimension=1536)
    })
    ```
    """
    ...

def i8_vector(dimension: builtins.int) -> FieldSpec:
    """
    Create a [FieldSpec](https://docs.topk.io/sdk/topk-py/schema#FieldSpec) type for `i8_vector` values.

    Example:

    ```python
    from topk_sdk.schema import i8_vector

    client.collections().create("books", schema={
        "title_embedding": i8_vector(dimension=1536)
    })
    ```
    """
    ...

def binary_vector(dimension: builtins.int) -> FieldSpec:
    """
    Create a [FieldSpec](https://docs.topk.io/sdk/topk-py/schema#FieldSpec) type for `binary_vector` values.

    Example:

    ```python
    from topk_sdk.schema import binary_vector

    client.collections().create("books", schema={
        "title_embedding": binary_vector(dimension=128)
    })
    ```
    """
    ...

def f32_sparse_vector() -> FieldSpec:
    """
    Create a [FieldSpec](https://docs.topk.io/sdk/topk-py/schema#FieldSpec) type for `f32_sparse_vector` values.

    Note: Sparse vectors use u32 dimension indices to support dictionaries of up to 2^32 - 1 terms.

    Example:

    ```python
    from topk_sdk.schema import f32_sparse_vector

    client.collections().create("books", schema={
        "sparse_field": f32_sparse_vector()
    })
    ```
    """
    ...

def u8_sparse_vector() -> FieldSpec:
    """
    Create a [FieldSpec](https://docs.topk.io/sdk/topk-py/schema#FieldSpec) type for `u8_sparse_vector` values.

    Note: Sparse vectors use u32 dimension indices to support dictionaries of up to 2^32 - 1 terms.

    Example:

    ```python
    from topk_sdk.schema import u8_sparse_vector

    client.collections().create("books", schema={
        "sparse_field": u8_sparse_vector()
    })
    ```
    """
    ...

def bytes() -> FieldSpec:
    """
    Create a [FieldSpec](https://docs.topk.io/sdk/topk-py/schema#FieldSpec) type for `bytes` values.

    Example:

    ```python
    from topk_sdk.schema import bytes

    client.collections().create("books", schema={
        "image": bytes()
    })
    ```
    """
    ...

def list(value_type: typing.Literal["text", "integer", "float"]) -> FieldSpec:
    """
    Create a [FieldSpec](https://docs.topk.io/sdk/topk-py/schema#FieldSpec) type for `list` values.

    Example:

    ```python
    from topk_sdk.schema import list

    client.collections().create("books", schema={
        "tags": list(value_type="text")
    })
    ```
    """
    ...

# indexes
def vector_index(
    metric: typing.Literal["cosine", "euclidean", "dot_product", "hamming"],
) -> FieldIndex:
    """
    Create a [FieldIndex](https://docs.topk.io/sdk/topk-py/schema#FieldSpec) type for `vector_index` values.

    Supported `metric`s:
        - `euclidean` (only dense vectors)
        - `cosine` (only dense vectors)
        - `dot_product` (dense and sparse vectors)
        - `hamming` (only binary vectors)

    Example:

    ```python
    from topk_sdk.schema import f32_vector, vector_index

    client.collections().create("books", schema={
        "title_embedding": f32_vector(dimension=1536).index(vector_index(metric="cosine"))
    })
    ```
    """
    ...

def keyword_index() -> FieldIndex:
    """
    Create a [FieldIndex](https://docs.topk.io/sdk/topk-py/schema#FieldIndex) type for `keyword_index` values.

    Example:

    ```python
    from topk_sdk.schema import text, keyword_index

    client.collections().create("books", schema={
        "title": text().index(keyword_index())
    })
    ```
    """
    ...

def semantic_index(model: str) -> FieldIndex:
    """
    Create a [FieldIndex](https://docs.topk.io/sdk/topk-py/schema#FieldIndex) type for `semantic_index` values.

    Supported `model`s:
        - `cohere/embed-english-v3`
        - `cohere/embed-multilingual-v3`
        - `cohere/embed-v4` (default)

    TopK supports the following embedding types for Cohere models:
        - `float32`
        - `uint8`
        - `binary`

    Example:

    ```python
    from topk_sdk.schema import text, semantic_index

    client.collections().create("books", schema={
        "title": text().index(semantic_index(model="cohere/embed-v4"))
    })
    ```
    """
    ...



from topk_sdk import data
from topk_sdk.schema import (
    binary_vector,
    f32_sparse_vector,
    f32_vector,
    i8_vector,
    int,
    keyword_index,
    semantic_index,
    text,
    u8_sparse_vector,
    u8_vector,
    vector_index,
    list,
)

from .. import ProjectContext, AsyncProjectContext


class books:
    @staticmethod
    def setup(ctx: ProjectContext):
        collection = ctx.client.collections().create(
            ctx.scope("books"),
            schema=books.schema(),
        )

        # Upsert sample books
        ctx.client.collection(collection.name).upsert(books.docs())

        return collection

    @staticmethod
    async def setup_async(ctx: AsyncProjectContext):
        collection = await ctx.client.collections().create(
            ctx.scope("books"),
            schema=books.schema(),
        )

        await ctx.client.collection(collection.name).upsert(books.docs())

        return collection

    @staticmethod
    def schema():
        return {
            "title": text().required().index(keyword_index()),
            "published_year": int().required(),
            "summary": text().required().index(keyword_index()),
            "summary_embedding": f32_vector(16)
            .required()
            .index(vector_index(metric="euclidean")),
            "nullable_embedding": f32_vector(16).index(
                vector_index(metric="euclidean")
            ),
            "scalar_embedding": u8_vector(16).index(vector_index(metric="euclidean")),
            "scalar_i8_embedding": i8_vector(16).index(
                vector_index(metric="euclidean")
            ),
            "binary_embedding": binary_vector(2).index(vector_index(metric="hamming")),
            "sparse_f32_embedding": f32_sparse_vector().index(
                vector_index(metric="dot_product")
            ),
            "sparse_u8_embedding": u8_sparse_vector().index(
                vector_index(metric="dot_product")
            ),
            "tags": list(value_type="text").index(keyword_index()),
            "codes": list(value_type="text"),
        }

    @staticmethod
    def docs():
        return [
            {
                "_id": "mockingbird",
                "title": "To Kill a Mockingbird",
                "published_year": 1960,
                "summary": "A young girl confronts racial injustice in the Deep South through the eyes of her lawyer father.",
                "summary_embedding": [1.0] * 16,
                "nullable_embedding": [1.0] * 16,
                "scalar_embedding": data.u8_vector([1] * 16),
                "scalar_i8_embedding": data.i8_vector([-100] * 16),
                "binary_embedding": data.binary_vector([0, 1]),
                "sparse_f32_embedding": data.f32_sparse_vector(
                    {0: 1.0, 1: 2.0, 2: 3.0}
                ),
                "sparse_u8_embedding": data.u8_sparse_vector({0: 1, 1: 2, 2: 3}),
                "tags": ["racism", "injustice", "girl", "father", "lawyer"],
                "codes": [
                    "ISBN 978-0-06-093546-7",
                    "ISBN 0-06-093546-4",
                    "LCCN 60007854",
                    "UPC 025192354670",
                ],
                "reprint_years": data.u32_list([1966, 1988, 1999, 2002, 2015]),
                "user_ratings": data.u32_list([9, 7, 10, 8]),
            },
            {
                "_id": "1984",
                "title": "1984",
                "published_year": 1949,
                "summary": "A totalitarian regime uses surveillance and mind control to oppress its citizens.",
                "summary_embedding": [2.0] * 16,
                "nullable_embedding": [2.0] * 16,
                "scalar_embedding": data.u8_vector([2] * 16),
                "scalar_i8_embedding": data.i8_vector([-50] * 16),
                "binary_embedding": data.binary_vector([0, 3]),
                "sparse_f32_embedding": data.f32_sparse_vector(
                    {2: 1.0, 3: 2.0, 4: 3.0}
                ),
                "sparse_u8_embedding": data.u8_sparse_vector({2: 1, 3: 2, 4: 3}),
                "tags": [
                    "dystopia",
                    "surveillance",
                    "totalitarianism",
                    "mind control",
                    "oppression",
                ],
                "codes": [
                    "1984",
                    "ISBN 978-0-452-28423-4",
                    "ISBN 0-452-28423-6",
                    "OCLC 70775700",
                    "DOI 10.1000/182",
                    "EAN 9780452284234",
                ],
                "reprint_years": data.u32_list(
                    [1950, 1954, 1956, 1961, 1984, 1990, 2003]
                ),
                "user_ratings": data.f32_list([5.0, 3.5, 4.5, 4.0, -1.0]),
            },
            {
                "_id": "pride",
                "title": "Pride and Prejudice",
                "published_year": 1813,
                "summary": "A witty exploration of love, social class, and marriage in 19th-century England.",
                "summary_embedding": [3.0] * 16,
                "scalar_i8_embedding": data.i8_vector([0] * 16),
                "sparse_f32_embedding": data.f32_sparse_vector(
                    {3: 1.0, 4: 2.0, 5: 3.0}
                ),
                "sparse_u8_embedding": data.u8_sparse_vector({3: 1, 4: 2, 5: 3}),
                "tags": ["pride", "love", "romance", "class", "marriage", "prejudice"],
                "codes": ["ISBN 978-0-14-143951-8", "OCLC 934546789"],
                "reprint_years": data.u32_list([1966, 1972, 1985]),
                "user_ratings": ["romantic", "classic", "good"],
            },
            {
                "_id": "gatsby",
                "title": "The Great Gatsby",
                "published_year": 1925,
                "summary": "A mysterious millionaire navigates love and wealth in the Roaring Twenties.",
                "summary_embedding": [4.0] * 16,
                "scalar_i8_embedding": data.i8_vector([50] * 16),
                "sparse_f32_embedding": data.f32_sparse_vector(
                    {4: 1.0, 5: 2.0, 6: 3.0}
                ),
                "sparse_u8_embedding": data.u8_sparse_vector({4: 1, 5: 2, 6: 3}),
                "tags": ["love", "romance", "wealth", "marriage"],
                "codes": [
                    "ISBN 978-0-7432-7356-5",
                    "ISBN 0-7432-7356-7",
                    "OCLC 60393320",
                    "ASIN B000FC0SIS",
                    "UPC 074327356709",
                    "LCCN 2002114274",
                ],
                "reprint_years": data.u32_list([1953, 1996, 2004]),
                "user_ratings": "good book",
            },
            {
                "_id": "catcher",
                "title": "The Catcher in the Rye",
                "published_year": 1951,
                "summary": "A rebellious teenager struggles with alienation and identity in mid-20th-century America.",
                "summary_embedding": [5.0] * 16,
                "nullable_embedding": [5.0] * 16,
                "scalar_embedding": data.u8_vector([5] * 16),
                "scalar_i8_embedding": data.i8_vector([100] * 16),
                "binary_embedding": data.binary_vector([0, 7]),
                "sparse_f32_embedding": data.f32_sparse_vector(
                    {5: 1.0, 6: 2.0, 7: 3.0}
                ),
                "sparse_u8_embedding": data.u8_sparse_vector({5: 1, 6: 2, 7: 3}),
                "tags": [
                    "alienation",
                    "identity",
                    "rebellion",
                    "mid-20th-century",
                    "america",
                ],
                "codes": ["ISBN 978-0-316-76917-4", "LCCN 51011564", "OCLC 287628"],
                "reprint_years": data.u32_list([1964, 1979, 1991, 2001, 2010]),
                "user_ratings": data.f32_list([4.5, 3.0, 3.5, 4.0, -1.0]),
            },
            {
                "_id": "moby",
                "title": "Moby-Dick",
                "published_year": 1851,
                "summary": "A sailor's obsessive quest to hunt a great white whale leads to tragic consequences.",
                "summary_embedding": [6.0] * 16,
                "sparse_f32_embedding": data.f32_sparse_vector(
                    {6: 1.0, 7: 2.0, 8: 3.0}
                ),
                "sparse_u8_embedding": data.u8_sparse_vector({6: 1, 7: 2, 8: 3}),
                "nullable_importance": 5.0,
                "tags": ["whale", "obsession", "tragedy", "sailing", "ocean"],
                "codes": data.string_list([]),
                "reprint_years": data.u32_list([]),
                "user_ratings": data.i32_list([-5, 2, -1, 1]),
            },
            {
                "_id": "hobbit",
                "title": "The Hobbit",
                "published_year": 1937,
                "summary": "A reluctant hobbit embarks on a quest to help a group of dwarves reclaim their mountain home.",
                "summary_embedding": [7.0] * 16,
                "sparse_f32_embedding": data.f32_sparse_vector(
                    {7: 1.0, 8: 2.0, 9: 3.0}
                ),
                "sparse_u8_embedding": data.u8_sparse_vector({7: 1, 8: 2, 9: 3}),
                "tags": ["hobbit", "dwarf", "quest", "home", "adventure"],
                "codes": data.string_list([]),
                "user_ratings": data.u32_list([5, 3, 4, 5, 2]),
            },
            {
                "_id": "harry",
                "title": "Harry Potter and the Sorcerer's Stone",
                "published_year": 1997,
                "summary": "A young wizard discovers his magical heritage and attends a school for witchcraft and wizardry.",
                "summary_embedding": [8.0] * 16,
                "nullable_embedding": [8.0] * 16,
                "scalar_embedding": data.u8_vector([8] * 16),
                "binary_embedding": data.binary_vector([0, 15]),
                "sparse_f32_embedding": data.f32_sparse_vector(
                    {8: 1.0, 9: 2.0, 10: 3.0}
                ),
                "sparse_u8_embedding": data.u8_sparse_vector({8: 1, 9: 2, 10: 3}),
                "tags": ["wizard", "magic", "sorcerer", "school", "witchcraft"],
                "codes": ["ISBN 978-0-439-70818-0", "UPC 043970818909"],
                "reprint_years": data.u32_list([1998, 1999, 2001, 2004, 2013, 2020]),
                "user_ratings": 10,
            },
            {
                "_id": "lotr",
                "title": "The Lord of the Rings: The Fellowship of the Ring",
                "published_year": 1954,
                "summary": "A group of unlikely heroes sets out to destroy a powerful, evil ring.",
                "summary_embedding": [9.0] * 16,
                "scalar_i8_embedding": data.i8_vector([-100] * 16),
                "sparse_f32_embedding": data.f32_sparse_vector(
                    {9: 1.0, 10: 2.0, 11: 3.0}
                ),
                "sparse_u8_embedding": data.u8_sparse_vector({9: 1, 10: 2, 11: 3}),
                "tags": ["lord of the rings", "fellowship", "magic", "wizard", "elves"],
                "codes": [
                    "ISBN 978-0-547-92821-0",
                    "ISBN 0-547-92821-2",
                    "OCLC 434394005",
                    "LCCN 2004558654",
                    "Barcode 0618346252",
                ],
                "user_ratings": ["epic", "legendary", "good"],
            },
            {
                "_id": "alchemist",
                "title": "The Alchemist",
                "published_year": 1988,
                "summary": "A shepherd boy journeys to fulfill his destiny and discover the meaning of life.",
                "summary_embedding": [10.0] * 16,
                "sparse_f32_embedding": data.f32_sparse_vector(
                    {10: 1.0, 11: 2.0, 12: 3.0}
                ),
                "sparse_u8_embedding": data.u8_sparse_vector({10: 1, 11: 2, 12: 3}),
                "tags": ["journey", "destiny", "meaning of life", "alchemy", "soul"],
                "codes": [
                    "ISBN 978-0-06-231500-7",
                    "ASIN 0062315005",
                    "OCLC 804616251",
                    "DOI 10.1234/alchemist",
                    "EAN 9780062315007",
                    "UPC 006231500719",
                    "LCCN 88675123",
                ],
                "reprint_years": data.u32_list([1993, 2005, 2014]),
                "user_ratings": data.u32_list([8, 10, 9, 7]),
            },
        ]


class semantic:
    @staticmethod
    def setup(ctx: ProjectContext):
        schema = {
            "title": text().required().index(semantic_index(model="dummy")),
            "summary": text().required().index(semantic_index(model="dummy")),
        }

        collection = ctx.client.collections().create(
            ctx.scope("semantic"),
            schema=schema,
        )

        # Upsert sample books
        ctx.client.collection(collection.name).upsert(semantic.docs())

        return collection

    @staticmethod
    def docs():
        return [
            {
                "_id": "mockingbird",
                "title": "To Kill a Mockingbird",
                "published_year": 1960,
                "summary": "A young girl confronts racial injustice in the Deep South through the eyes of her lawyer father.",
                "nullable_importance": 2.0,
            },
            {
                "_id": "1984",
                "title": "1984",
                "published_year": 1949,
                "summary": "A totalitarian regime uses surveillance and mind control to oppress its citizens.",
            },
            {
                "_id": "pride",
                "title": "Pride and Prejudice",
                "published_year": 1813,
                "summary": "A witty exploration of love, social class, and marriage in 19th-century England.",
            },
            {
                "_id": "gatsby",
                "title": "The Great Gatsby",
                "published_year": 1925,
                "summary": "A mysterious millionaire navigates love and wealth in the Roaring Twenties.",
            },
            {
                "_id": "catcher",
                "title": "The Catcher in the Rye",
                "published_year": 1951,
                "summary": "A rebellious teenager struggles with alienation and identity in mid-20th-century America.",
            },
            {
                "_id": "moby",
                "title": "Moby-Dick",
                "published_year": 1851,
                "summary": "A sailor's obsessive quest to hunt a great white whale leads to tragic consequences.",
            },
            {
                "_id": "hobbit",
                "title": "The Hobbit",
                "published_year": 1937,
                "summary": "A reluctant hobbit embarks on a quest to help a group of dwarves reclaim their mountain home.",
            },
            {
                "_id": "harry",
                "title": "Harry Potter and the Sorcerer's Stone",
                "published_year": 1997,
                "summary": "A young wizard discovers his magical heritage and attends a school for witchcraft and wizardry.",
            },
            {
                "_id": "lotr",
                "title": "The Lord of the Rings: The Fellowship of the Ring",
                "published_year": 1954,
                "summary": "A group of unlikely heroes sets out to destroy a powerful, evil ring.",
            },
            {
                "_id": "alchemist",
                "title": "The Alchemist",
                "published_year": 1988,
                "summary": "A shepherd boy journeys to fulfill his destiny and discover the meaning of life.",
            },
        ]



import pytest
from topk_sdk import error
from topk_sdk.query import field, fn, match, select

from . import ProjectContext
from .utils import dataset, doc_ids, doc_fields


def test_query_bare_limit(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(select().limit(100))

    assert len(result) == 10
    expected_ids = {str(doc["_id"]) for doc in dataset.books.docs()}
    assert doc_ids(result) == expected_ids


def test_query_limit_select_filter(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(
            _id=field("_id"),
            summary=field("summary"),
            is_recent=field("published_year") > 1980,
        )
        .filter(field("_id") <= "hobbit")
        .limit(3)
    )

    assert len(result) == 3
    assert doc_fields(result) == {"_id", "summary", "is_recent"}

    expected_ids = {"1984", "alchemist", "catcher", "gatsby", "harry", "hobbit"}
    for doc in result:
        assert doc["_id"] in expected_ids


def test_query_limit_with_bm25(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(bm25=fn.bm25_score()).filter(match("quest", all=True)).limit(10)
    )

    assert len(result) == 2
    assert doc_fields(result) == {"_id", "bm25"}
    assert {doc["_id"] for doc in result} == {"moby", "hobbit"}


def test_query_limit_vector_distance(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    # skip_refine is implicitly true
    result_limit = ctx.client.collection(collection.name).query(
        select(title=field("title"))
        .select(summary_distance=fn.vector_distance("summary_embedding", [2.0] * 16))
        .limit(100)
    )

    # explicitly set skip_refine to true
    result_topk = ctx.client.collection(collection.name).query(
        select(title=field("title"))
        .select(
            summary_distance=fn.vector_distance("summary_embedding", [2.0] * 16, True)
        )
        .topk(field("summary_distance"), 100, True)
    )

    docs_limit = {doc["_id"]: doc for doc in result_limit}
    docs_topk = {doc["_id"]: doc for doc in result_topk}

    assert doc_fields(result_limit) == {"_id", "title", "summary_distance"}
    assert doc_fields(result_topk) == {"_id", "title", "summary_distance"}
    assert len(docs_limit) == len(docs_topk)

    # vector distance from limit should be the same as the vector distance from topk with skip_refine true
    assert docs_limit == docs_topk


def test_query_invalid_collectors(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    invalid_queries = [
        # topk + limit - multiple collectors
        select(title=field("title"))
        .topk(field("published_year"), 100, True)
        .limit(100),
        # limit + count - multiple collectors
        select(title=field("title")).limit(100).count(),
        # no collector
        select(title=field("title")).sort(field("published_year"), True),
        # multiple sorts
        select(title=field("title"))
        .sort(field("published_year"), True)
        .sort(field("published_year"), False),
        # topk + sort - effectively multiple sorts
        select(title=field("title"))
        .topk(field("published_year"), 100, True)
        .sort(field("published_year"), True),
    ]

    for q in invalid_queries:
        with pytest.raises(error.InvalidArgumentError):
            ctx.client.collection(collection.name).query(q)



import pytest
from topk_sdk.data import bytes


def test_bytes_from_list():
    result = bytes([0, 1, 255, 128])
    assert result is not None
    assert str(result) == str(b"\x00\x01\xff\x80")


def test_bytes_from_bytes():
    result = bytes(b"\x00\x01\xff\x80")
    assert result is not None
    assert str(result) == str(b"\x00\x01\xff\x80")


def test_bytes_invalid_input():
    with pytest.raises(TypeError, match="Expected bytes or list\\[int\\]"):
        bytes("not a valid input")  # type: ignore

    with pytest.raises(TypeError, match="Expected bytes or list\\[int\\]"):
        bytes(123)  # type: ignore


def test_bytes_empty():
    result1 = bytes([])
    result2 = bytes(b"")
    assert result1 is not None
    assert result2 is not None
    assert str(result1) == "b''"
    assert str(result2) == "b''"


def test_bytes_with_invalid_list_values():
    with pytest.raises(
        TypeError,
        match="Expected list\\[int\\] with values in range \\[0, 255\\]",
    ):
        bytes([0, 1, "invalid", 3])  # type: ignore

    with pytest.raises(
        TypeError,
        match="Expected list\\[int\\] with values in range \\[0, 255\\]",
    ):
        bytes([0, 1, 256, 3])

    with pytest.raises(
        TypeError,
        match="Expected list\\[int\\] with values in range \\[0, 255\\]",
    ):
        bytes([0, -1, 2, 3])



import pytest
import pytest_asyncio


@pytest.fixture
def ctx():
    from . import new_project_context

    ctx = new_project_context()

    yield ctx

    ctx.cleanup()


@pytest_asyncio.fixture
async def async_ctx():
    from . import new_async_project_context

    ctx = new_async_project_context()

    yield ctx

    await ctx.cleanup()



import builtins
import typing
from enum import Enum

from . import query, schema

class Client:
    """
    Client for interacting with the TopK API. For available regions see https://docs.topk.io/regions
    """

    def __init__(
        self,
        api_key: builtins.str,
        region: builtins.str,
        host: builtins.str = "topk.io",
        https: builtins.bool = True,
        retry_config: typing.Optional[RetryConfig | dict[builtins.str, typing.Any]] = None,
    ) -> None: ...
    def collection(self, collection: builtins.str) -> CollectionClient:
        """
        Get a client for managing data operations on a specific collection such as querying, upserting, and deleting documents.
        """
        ...
    def collections(self) -> CollectionsClient:
        """
        Get a client for managing collections.
        """
        ...

class AsyncClient:
    """
    Async client for interacting with the TopK API. For available regions see https://docs.topk.io/regions
    """

    def __init__(
        self,
        api_key: builtins.str,
        region: builtins.str,
        host: builtins.str = "topk.io",
        https: builtins.bool = True,
        retry_config: typing.Optional[RetryConfig | dict[builtins.str, typing.Any]] = None,
    ) -> None: ...
    def collection(self, collection: builtins.str) -> AsyncCollectionClient:
        """Get an async client for a specific collection."""
        ...
    def collections(self) -> AsyncCollectionsClient:
        """Get an async client for managing collections."""
        ...

class CollectionClient:
    """
    Synchronous client for collection operations.
    """

    def get(
        self,
        ids: typing.Sequence[builtins.str],
        fields: typing.Optional[typing.Sequence[builtins.str]] = None,
        lsn: typing.Optional[builtins.str] = None,
        consistency: typing.Optional[ConsistencyLevel] = None,
    ) -> builtins.dict[builtins.str, builtins.dict[builtins.str, typing.Any]]:
        """
        Get documents by their IDs.
        """
        ...
    def count(
        self,
        lsn: typing.Optional[builtins.str] = None,
        consistency: typing.Optional[ConsistencyLevel] = None,
    ) -> builtins.int:
        """
        Get the count of documents in the collection.
        """
        ...
    def query(
        self,
        query: query.Query,
        lsn: typing.Optional[builtins.str] = None,
        consistency: typing.Optional[ConsistencyLevel] = None,
    ) -> builtins.list[builtins.dict[builtins.str, typing.Any]]:
        """
        Execute a query against the collection.
        """
        ...
    def upsert(
        self, documents: typing.Sequence[typing.Mapping[builtins.str, typing.Any]]
    ) -> builtins.str:
        """
        Insert or update documents in the collection.
        """
        ...
    def delete(self, expr: typing.Union[typing.Sequence[builtins.str], query.LogicalExpr]) -> builtins.str:
        """
        Delete documents by their IDs or using a filter expression.

        **Example:**

        Delete documents by their IDs:
        ```python
        client.collection("books").delete(["id_1", "id_2"])
        ```

        Delete documents by a filter expression:
        ```python
        from topk_sdk.query import field

        client.collection("books").delete(field("published_year").gt(1997))
        ```
        """
        ...

class AsyncCollectionClient:
    """
    Asynchronous client for collection operations.
    """

    def get(
        self,
        ids: typing.Sequence[builtins.str],
        fields: typing.Optional[typing.Sequence[builtins.str]] = None,
        lsn: typing.Optional[builtins.str] = None,
        consistency: typing.Optional[ConsistencyLevel] = None,
    ) -> typing.Awaitable[
        builtins.dict[builtins.str, builtins.dict[builtins.str, typing.Any]]
    ]:
        """
        Get documents by their IDs asynchronously.
        """
        ...
    def count(
        self,
        lsn: typing.Optional[builtins.str] = None,
        consistency: typing.Optional[ConsistencyLevel] = None,
    ) -> typing.Awaitable[builtins.int]:
        """
        Get the count of documents in the collection asynchronously.
        """
        ...
    def query(
        self,
        query: query.Query,
        lsn: typing.Optional[builtins.str] = None,
        consistency: typing.Optional[ConsistencyLevel] = None,
    ) -> typing.Awaitable[builtins.list[builtins.dict[builtins.str, typing.Any]]]:
        """
        Execute a query against the collection asynchronously.
        """
        ...
    def upsert(
        self, documents: typing.Sequence[typing.Mapping[builtins.str, typing.Any]]
    ) -> typing.Awaitable[builtins.str]:
        """
        Insert or update documents in the collection asynchronously.
        """
        ...
    def delete(
        self, expr: typing.Union[typing.Sequence[builtins.str], query.LogicalExpr]
    ) -> typing.Awaitable[builtins.str]:
        """
        Delete documents by their IDs or using a filter expression asynchronously.

        **Example:**

        Delete documents by their IDs:
        ```python
        await client.collection("books").delete(["id_1", "id_2"])
        ```

        Delete documents by a filter expression:
        ```python
        from topk_sdk.query import field

        await client.collection("books").delete(field("published_year").gt(1997))
        ```
        """
        ...

class Collection:
    """
    Represents a collection in the TopK system.
    """

    name: builtins.str
    org_id: builtins.str
    project_id: builtins.str
    region: builtins.str
    schema: builtins.dict[builtins.str, schema.FieldSpec]

class CollectionsClient:
    """
    Synchronous client for managing collections.
    """

    def get(self, collection_name: builtins.str) -> Collection:
        """
        Get information about a specific collection.
        """
        ...
    def list(self) -> builtins.list[Collection]:
        """
        List all collections.
        """
        ...
    def create(
        self,
        collection_name: builtins.str,
        schema: typing.Mapping[builtins.str, schema.FieldSpec],
    ) -> Collection:
        """
        Create a new collection with the specified schema.
        """
        ...
    def delete(self, collection_name: builtins.str) -> None:
        """
        Delete a collection.
        """
        ...

class AsyncCollectionsClient:
    """
    Asynchronous client for managing collections.
    """

    def get(self, collection_name: builtins.str) -> typing.Awaitable[Collection]:
        """
        Get information about a specific collection asynchronously.
        """
        ...
    def list(self) -> typing.Awaitable[builtins.list[Collection]]:
        """
        List all collections asynchronously.
        """
        ...
    def create(
        self,
        collection_name: builtins.str,
        schema: typing.Mapping[builtins.str, schema.FieldSpec],
    ) -> typing.Awaitable[Collection]:
        """
        Create a new collection with the specified schema asynchronously.
        """
        ...
    def delete(self, collection_name: builtins.str) -> typing.Awaitable[None]:
        """
        Delete a collection asynchronously.
        """
        ...

class ConsistencyLevel(Enum):
    """
    Enumeration of consistency levels for operations.
    """

    Indexed = "indexed"
    Strong = "strong"

class RetryConfig:
    """
    Configuration for retry behavior.

    By default, retries occur in two situations:
    1. When the server requests the client to reduce its request rate, resulting in a [SlowDownError](https://docs.topk.io/sdk/topk-py/error#slowdownerror).
    2. When using the `query(..., lsn=N)` to wait for writes to be available.
    """

    def __init__(
        self,
        max_retries: typing.Optional[builtins.int] = None,
        timeout: typing.Optional[builtins.int] = None,
        backoff: typing.Optional[BackoffConfig] = None,
    ) -> None: ...

    max_retries: typing.Annotated[typing.Optional[builtins.int], "Maximum number of retries to attempt. Default is 3 retries."]
    """
    Maximum number of retries to attempt.
    Default is 3 retries.
    """
    timeout: typing.Annotated[typing.Optional[builtins.int], "The total timetout for the retry chain in milliseconds. Default is 30,000 milliseconds (30 seconds)"]
    """
    The total timeout for the retry chain in milliseconds.
    Default is 30,000 milliseconds (30 seconds).
    """
    backoff: typing.Annotated[typing.Optional[BackoffConfig], "The backoff configuration for the client."]
    """
    The backoff configuration for the client.
    """

class BackoffConfig:
    """
    Configuration for backoff behavior in retries.
    """

    def __init__(
        self,
        base: typing.Optional[builtins.int] = None,
        init_backoff: typing.Optional[builtins.int] = None,
        max_backoff: typing.Optional[builtins.int] = None,
    ) -> None: ...

    base: typing.Annotated[typing.Optional[builtins.int], "The base for the backoff. Default is 2x backoff."]
    """
    The base for the backoff. Default is 2x backoff.
    """
    init_backoff: typing.Annotated[typing.Optional[builtins.int], "The initial backoff in milliseconds. Default is 100 milliseconds."]
    """
    The initial backoff in milliseconds.
    Default is 100 milliseconds.
    """
    max_backoff: typing.Annotated[typing.Optional[builtins.int], "The maximum backoff in milliseconds. Default is 10,000 milliseconds (10 seconds)."]
    """
    The maximum backoff in milliseconds.
    Default is 10,000 milliseconds (10 seconds).
    """



import builtins
import typing

class List:
    """
    *Internal*

    Instances of the `List` class are used to represent lists of values in TopK.
    Usually created using data constructors such as [`f32_list()`](#f32-list), [`i32_list()`](#i32-list), etc.
    """

    ...

class SparseVector:
    """
    *Internal*

    Instances of the `SparseVector` class are used to represent sparse vectors in TopK.
    Usually created using data constructors such as [`f32_sparse_vector()`](#f32-sparse-vector) or [`u8_sparse_vector()`](#u8-sparse-vector).
    """

    ...

def f32_vector(data: builtins.list[float]) -> List:
    """
    Create a [List](https://docs.topk.io/sdk/topk-py/data#List) type containing a 32-bit float vector. This function is an alias for [f32_list()](https://docs.topk.io/sdk/topk-py/data#f32-list).

    Example:

    ```python
    from topk_sdk.data import f32_vector

    f32_vector([0.12, 0.67, 0.82, 0.53])
    ```
    """
    ...

def u8_vector(data: builtins.list[int]) -> List:
    """
    Create a [List](https://docs.topk.io/sdk/topk-py/data#List) type containing an 8-bit unsigned integer vector. This function is an alias for [u8_list()](https://docs.topk.io/sdk/topk-py/data#u8-list).

    Example:

    ```python
    from topk_sdk.data import u8_vector

    u8_vector([0, 255, 1, 2, 3])
    ```
    """
    ...

def i8_vector(data: builtins.list[int]) -> List:
    """
    Create a [List](https://docs.topk.io/sdk/topk-py/data#List) type containing an 8-bit signed integer vector.

    Example:

    ```python
    from topk_sdk.data import i8_vector

    i8_vector([-128, 127, -1, 0, 1])
    ```
    """
    ...

def binary_vector(data: builtins.list[int]) -> List:
    """
    Create a [List](https://docs.topk.io/sdk/topk-py/data#List) type containing a binary vector.

    Example:

    ```python
    from topk_sdk.data import binary_vector

    binary_vector([0, 1, 1, 0])
    ```
    """
    ...

def f32_sparse_vector(data: builtins.dict[int, float]) -> SparseVector:
    """
    Create a [SparseVector](https://docs.topk.io/sdk/topk-py/data#SparseVector) type containing a 32-bit float sparse vector.

    Example:

    ```python
    from topk_sdk.data import f32_sparse_vector

    f32_sparse_vector({0: 0.12, 6: 0.67, 17: 0.82, 97: 0.53})
    ```
    """
    ...

def u8_sparse_vector(data: builtins.dict[int, int]) -> SparseVector:
    """
    Create a [SparseVector](https://docs.topk.io/sdk/topk-py/data#SparseVector) type containing an 8-bit unsigned integer sparse vector.

    Example:

    ```python
    from topk_sdk.data import u8_sparse_vector

    u8_sparse_vector({0: 12, 6: 67, 17: 82, 97: 53})
    ```
    """
    ...

def bytes(data: typing.Union[builtins.list[int], builtins.bytes]) -> List:
    """
    Create a [List](https://docs.topk.io/sdk/topk-py/data#List) type containing bytes data.

    Example:

    ```python
    from topk_sdk.data import bytes

    bytes([0, 1, 1, 0])
    ```
    """
    ...

def u32_list(data: builtins.list[int]) -> List:
    """
    Create a [List](https://docs.topk.io/sdk/topk-py/data#List) type containing a list of 32-bit unsigned integers.

    Example:

    ```python
    from topk_sdk.data import u32_list

    u32_list([0, 1, 2, 3])
    ```
    """
    ...

def i32_list(data: builtins.list[int]) -> List:
    """
    Create a [List](https://docs.topk.io/sdk/topk-py/data#List) type containing a list of 32-bit signed integers.

    Example:

    ```python
    from topk_sdk.data import i32_list

    i32_list([0, 1, 2, 3])
    ```
    """
    ...

def i64_list(data: builtins.list[int]) -> List:
    """
    Create a [List](https://docs.topk.io/sdk/topk-py/data#List) type containing a list of 64-bit signed integers.

    Example:

    ```python
    from topk_sdk.data import i64_list

    i64_list([0, 1, 2, 3])
    ```
    """
    ...

def f32_list(data: builtins.list[float]) -> List:
    """
    Create a [List](https://docs.topk.io/sdk/topk-py/data#List) type containing a list of 32-bit floating point numbers.

    Example:

    ```python
    from topk_sdk.data import f32_list

    f32_list([0.12, 0.67, 0.82, 0.53])
    ```
    """
    ...

def f64_list(data: builtins.list[float]) -> List:
    """
    Create a [List](https://docs.topk.io/sdk/topk-py/data#List) type containing a list of 64-bit floating point numbers.

    Example:

    ```python
    from topk_sdk.data import f64_list

    f64_list([0.12, 0.67, 0.82, 0.53])
    ```
    """
    ...

def string_list(data: builtins.list[str]) -> List:
    """
    Create a [List](https://docs.topk.io/sdk/topk-py/data#List) type containing a list of strings.

    Example:

    ```python
    from topk_sdk.data import string_list

    string_list(["foo", "bar", "baz"])
    ```
    """
    ...



import pytest
from topk_sdk.query import field, literal


def test_query_expr_with_flexible_expr():
    assert (field("a") + 1)._expr_eq(field("a") + literal(1))
    assert (1 + field("a"))._expr_eq(field("a") + literal(1))

    assert (field("a") - 1)._expr_eq(field("a") - literal(1))
    assert (1 - field("a"))._expr_eq(literal(1) - field("a"))

    assert (field("a") * 1)._expr_eq(field("a") * literal(1))
    assert (1 * field("a"))._expr_eq(field("a") * literal(1))

    assert (field("a") / 1)._expr_eq(field("a") / literal(1))
    assert (1 / field("a"))._expr_eq(literal(1) / field("a"))

    assert (field("a") & True)._expr_eq(field("a") & literal(True))
    assert (True & field("a"))._expr_eq(field("a") & literal(True))

    assert (field("a") | False)._expr_eq(field("a") | literal(False))
    assert (False | field("a"))._expr_eq(field("a") | literal(False))

    assert (field("a") & True)._expr_eq(field("a").and_(True))
    assert (True & field("a"))._expr_eq(field("a").and_(True))

    assert (field("a") | False)._expr_eq(field("a").or_(False))
    assert (False | field("a"))._expr_eq(field("a").or_(False))

    assert (field("a") & field("b"))._expr_eq(field("a").and_(field("b")))
    assert (field("a") | field("b"))._expr_eq(field("a").or_(field("b")))


def test_comparison_operators():
    assert (field("a") == 1)._expr_eq(field("a") == literal(1))
    assert (1 == field("a"))._expr_eq(field("a") == literal(1))  # type: ignore

    assert (field("a") != 1)._expr_eq(field("a") != literal(1))
    assert (1 != field("a"))._expr_eq(field("a") != literal(1))  # type: ignore

    assert (field("a") < 1)._expr_eq(field("a") < literal(1))
    assert (1 > field("a"))._expr_eq(field("a") < literal(1))

    assert (field("a") <= 1)._expr_eq(field("a") <= literal(1))
    assert (1 >= field("a"))._expr_eq(field("a") <= literal(1))

    assert (field("a") > 1)._expr_eq(field("a") > literal(1))
    assert (1 < field("a"))._expr_eq(field("a") > literal(1))

    assert (field("a") >= 1)._expr_eq(field("a") >= literal(1))
    assert (1 <= field("a"))._expr_eq(field("a") >= literal(1))


def test_query_expr_eq():
    assert (literal("a") + literal("b"))._expr_eq(literal("a") + literal("b"))
    assert not (literal("a")._expr_eq(literal("b")))
    assert field("a")._expr_eq(field("a"))
    assert not (field("a")._expr_eq(field("b")))
    assert not (field("a")._expr_eq(literal("a")))
    assert not (literal("a")._expr_eq(field("a")))


def test_query_literal():
    field("foo").eq(literal(1))
    field("foo").eq(1)

    field("foo").ne(literal(1))
    field("foo").ne(1)


def test_invalid():
    with pytest.raises(TypeError):
        literal(1) + "string"  # type: ignore

    with pytest.raises(TypeError):
        field("a") & 1  # type: ignore

    with pytest.raises(TypeError):
        field("a") | 1  # type: ignore

    with pytest.raises(TypeError):
        field("a") + None  # type: ignore

    with pytest.raises(TypeError):
        field("a") + "string"  # type: ignore

    with pytest.raises(TypeError):
        field("a") + [1, 2, 3]  # type: ignore

    with pytest.raises(TypeError):
        field("a") + {"a": 1}  # type: ignore


def test_invalid_truthiness():
    error_msg = "Using `and` or `or` keywords with Logical expressions is not supported. Please use `&` or `|` instead."

    # `and`
    with pytest.raises(TypeError) as e:
        field("a") == 1 and field("b") == 2
    assert error_msg in str(e.value)

    # `or`
    with pytest.raises(TypeError) as e:
        field("a") == 1 or field("b") == 2
    assert error_msg in str(e.value)



import pytest
from topk_sdk import error
from topk_sdk.query import field, filter, fn, match, select

from . import ProjectContext
from .utils import dataset, doc_ids


def test_query_text_filter_single_term_disjunctive(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(match("love", field="summary")).topk(field("published_year"), 100, True)
    )

    assert doc_ids(result) == {"pride", "gatsby"}


def test_query_text_filter_single_term_conjunctive(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(match("love", field="summary")).topk(field("published_year"), 100, True)
    )

    assert doc_ids(result) == {"gatsby", "pride"}


def test_query_text_filter_two_terms_disjunctive(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(match("LOVE", "summary") | match("ring", "title")).topk(
            field("published_year"), 100, True
        )
    )

    assert doc_ids(result) == {"pride", "gatsby", "lotr"}


def test_query_text_filter_two_terms_conjunctive(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(match("LOVE", field="summary") & match("class", field="summary")).topk(
            field("published_year"), 100, True
        )
    )

    assert doc_ids(result) == {"pride"}


def test_query_text_filter_stop_word(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(match("the", field="summary")).topk(field("published_year"), 100, True)
    )

    assert len(result) == 0


def test_query_select_bm25_without_text_queries(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    with pytest.raises(error.InvalidArgumentError):
        ctx.client.collection(collection.name).query(
            select(bm25_score=fn.bm25_score())
            .filter(field("_id") == "pride")
            .topk(field("bm25_score"), 100, True)
        )


def test_query_text_matches_single_term(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    for match_expr in [
        filter(field("summary").match_any("love")),
        filter(field("summary").match_all("love")),
    ]:
        result = ctx.client.collection(collection.name).query(
            match_expr.topk(field("published_year"), 100, True)
        )

        assert doc_ids(result) == {"pride", "gatsby"}


def test_query_text_match_all_two_terms(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(field("summary").match_all("love class")).topk(
            field("published_year"), 100, True
        )
    )

    assert doc_ids(result) == {"pride"}


def test_query_text_match_all_two_terms_tokenized(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(field("tags").match_all(["love", "class"])).topk(
            field("published_year"), 100, True
        )
    )

    assert doc_ids(result) == {"pride"}


def test_query_text_match_any_two_terms(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(field("summary").match_any("love ring")).topk(
            field("published_year"), 100, True
        )
    )

    assert doc_ids(result) == {"pride", "gatsby", "lotr"}


def test_query_text_match_any_two_terms_tokenized(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(field("tags").match_any(["love", "elves"])).topk(
            field("published_year"), 100, True
        )
    )

    assert doc_ids(result) == {"pride", "gatsby", "lotr"}


def test_query_text_matches_with_logical_expr(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        filter(
            (field("summary").match_all("love class"))
            | (field("published_year") == 1925)
        ).topk(field("published_year"), 10, True)
    )

    assert doc_ids(result) == {"pride", "gatsby"}


def test_query_text_matches_on_invalid_field(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    with pytest.raises(error.InvalidArgumentError):
        ctx.client.collection(collection.name).query(
            filter(field("published_year").match_all("love class")).count()
        )


def test_invalid_truthiness():
    error_msg = "Using `and` or `or` keywords with Text expressions is not supported. Please use `&` or `|` instead."

    # `and`
    with pytest.raises(TypeError) as e:
        match("foo") and match("bar")
    assert error_msg in str(e.value)

    # `or`
    with pytest.raises(TypeError) as e:
        match("foo") or match("bar")
    assert error_msg in str(e.value)



"""Parser module for extracting information from Python AST nodes."""

import ast
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional


@dataclass
class TypeAnnotation:
    """Represents a type annotation."""

    name: str
    is_optional: bool = False
    is_generic: bool = False
    generic_args: List["TypeAnnotation"] = None  # type: ignore

    def __post_init__(self):
        if self.generic_args is None:  # type: ignore
            self.generic_args = []


@dataclass
class Parameter:
    """Represents a function parameter."""

    name: str
    type_annotation: Optional[TypeAnnotation]
    default_value: Optional[str] = None
    is_required: bool = True


@dataclass
class Method:
    """Represents a class method."""

    name: str
    parameters: List[Parameter]
    return_type: Optional[TypeAnnotation]
    docstring: Optional[str] = None
    is_constructor: bool = False


@dataclass
class Property:
    """Represents a class property/attribute."""

    name: str
    type_annotation: TypeAnnotation
    docstring: Optional[str] = None


@dataclass
class EnumValue:
    """Represents an enum value."""

    name: str
    value: str


@dataclass
class Class:
    """Represents a class definition."""

    name: str
    docstring: Optional[str]
    is_enum: bool = False
    methods: List[Method] | None = None
    properties: List[Property] | None = None
    enum_values: List[EnumValue] | None = None

    def __post_init__(self):
        if self.methods is None:
            self.methods = []
        if self.properties is None:
            self.properties = []
        if self.enum_values is None:
            self.enum_values = []


@dataclass
class Function:
    """Represents a standalone function."""

    name: str
    parameters: List[Parameter]
    return_type: Optional[TypeAnnotation]
    docstring: Optional[str] = None


@dataclass
class TypeAlias:
    """Represents a type alias definition."""

    name: str
    type_annotation: TypeAnnotation
    docstring: Optional[str] = None


@dataclass
class Module:
    """Represents a parsed module."""

    file_path: Path
    classes: List[Class]
    functions: List[Function]
    type_aliases: List[TypeAlias] | None = None

    def __post_init__(self):
        if self.type_aliases is None:
            self.type_aliases = []


def format_type_annotation(annotation) -> TypeAnnotation:
    """Convert AST type annotation to TypeAnnotation object."""
    if annotation is None:
        return TypeAnnotation("None")

    if isinstance(annotation, ast.Name):
        # Clean up builtins references
        if annotation.id.startswith("builtins."):
            name = annotation.id.replace("builtins.", "")
        else:
            name = annotation.id
        return TypeAnnotation(name)

    elif isinstance(annotation, ast.Constant):
        return TypeAnnotation(repr(annotation.value))

    elif isinstance(annotation, ast.Attribute):
        # Clean up builtins references
        if isinstance(annotation.value, ast.Name):
            if annotation.value.id == "builtins":
                name = annotation.attr
            elif annotation.value.id == "typing":
                name = annotation.attr
            else:
                name = f"{annotation.value.id}.{annotation.attr}"
        else:
            # Handle nested attributes recursively
            value_name = format_type_annotation(annotation.value).name
            name = f"{value_name}.{annotation.attr}"
        return TypeAnnotation(name)

    elif isinstance(annotation, ast.Subscript):
        # Handle things like List[str], Optional[int], etc.
        if isinstance(annotation.value, ast.Name):
            base = annotation.value.id
        elif isinstance(annotation.value, ast.Attribute):
            # Handle nested attributes like typing.Union, builtins.str, etc.
            if isinstance(annotation.value.value, ast.Name):
                if annotation.value.value.id == "builtins":
                    base = annotation.value.attr
                elif annotation.value.value.id == "typing":
                    base = annotation.value.attr
                else:
                    base = f"{annotation.value.value.id}.{annotation.value.attr}"
            else:
                # Handle deeper nesting like typing.Union
                base = f"{annotation.value.value.id}.{annotation.value.attr}"  # type: ignore
        else:
            base = "Unknown"

        is_optional = base == "Optional"

        if isinstance(annotation.slice, ast.Tuple):
            args = [format_type_annotation(el) for el in annotation.slice.elts]
            return TypeAnnotation(
                base, is_optional=is_optional, is_generic=True, generic_args=args
            )
        else:
            arg = format_type_annotation(annotation.slice)
            return TypeAnnotation(
                base, is_optional=is_optional, is_generic=True, generic_args=[arg]
            )

    elif isinstance(annotation, ast.BinOp) and isinstance(annotation.op, ast.BitOr):
        # Handle union types like A | B (Python 3.10+)
        left = format_type_annotation(annotation.left)
        right = format_type_annotation(annotation.right)
        return TypeAnnotation("Union", is_generic=True, generic_args=[left, right])

    else:
        return TypeAnnotation("Unknown")


def extract_docstring(node) -> Optional[str]:
    """Extract docstring from a node."""
    if (
        isinstance(node, (ast.FunctionDef, ast.ClassDef))
        and node.body
        and isinstance(node.body[0], ast.Expr)
        and isinstance(node.body[0].value, ast.Constant)
        and isinstance(node.body[0].value.value, str)
    ):
        return node.body[0].value.value.strip()
    return None


def parse_parameter(arg: ast.arg, default_value=None) -> Parameter:
    """Parse a function parameter."""
    param_name = arg.arg
    type_annotation = None

    if arg.annotation:
        type_annotation = format_type_annotation(arg.annotation)

    is_required = default_value is None

    return Parameter(
        name=param_name,
        type_annotation=type_annotation,
        default_value=default_value,  # type: ignore
        is_required=is_required,
    )


def parse_method(method_node: ast.FunctionDef) -> Method | None:
    """Parse a method definition."""
    parameters: List[Parameter] = []
    defaults = method_node.args.defaults
    # default_args = method_node.args.args[-len(defaults) :] if defaults else []

    if method_node.name.startswith("__") and not method_node.name == "__init__":
        return None

    # skip _expr_eq method
    if method_node.name == "_expr_eq":
        return None

    for i, arg in enumerate(method_node.args.args):
        default_val = None
        if i >= len(method_node.args.args) - len(defaults):
            default_idx = i - (len(method_node.args.args) - len(defaults))
            if default_idx < len(defaults):
                default = defaults[default_idx]
                if isinstance(default, ast.Constant):
                    if isinstance(default.value, str):
                        default_val = f'"{default.value}"'
                    else:
                        default_val = repr(default.value)
                elif isinstance(default, ast.Constant):
                    default_val = repr(default.value)
                elif isinstance(default, ast.Str):  # type: ignore Python 3.7 compatibility
                    default_val = f'"{default.s}"'  # type: ignore
                elif isinstance(default, ast.Name):
                    default_val = default.id
                else:
                    default_val = "..."

        param = parse_parameter(arg, default_val)
        parameters.append(param)

    return_type = None
    if method_node.returns:
        return_type = format_type_annotation(method_node.returns)

    docstring = extract_docstring(method_node)

    return Method(
        name=method_node.name,
        parameters=parameters,
        return_type=return_type,
        docstring=docstring,
        is_constructor=(method_node.name == "__init__"),
    )


def parse_class(class_node: ast.ClassDef) -> Class:
    """Parse a class definition."""
    name = class_node.name
    docstring = extract_docstring(class_node)

    # Check if it's an enum
    is_enum = any(
        isinstance(sub, ast.Assign)
        and isinstance(sub.value, ast.Constant)
        and isinstance(sub.value.value, str)
        for sub in class_node.body
    )

    if is_enum:
        enum_values: List[EnumValue] = []
        for sub_node in class_node.body:
            if isinstance(sub_node, ast.Assign) and len(sub_node.targets) == 1:
                if isinstance(sub_node.targets[0], ast.Name) and isinstance(
                    sub_node.value, ast.Constant
                ):
                    enum_values.append(
                        EnumValue(
                            name=sub_node.targets[0].id,
                            value=sub_node.value.value,  # type: ignore
                        )
                    )
        return Class(
            name=name, docstring=docstring, is_enum=True, enum_values=enum_values
        )

    # Check if it's a data class (has attributes but no methods)
    methods = [sub for sub in class_node.body if isinstance(sub, ast.FunctionDef)]
    attributes = [sub for sub in class_node.body if isinstance(sub, ast.AnnAssign)]

    # Parse class properties
    properties: List[Property] = []
    for attr in attributes:
        if isinstance(attr.target, ast.Name):
            type_annotation = (
                format_type_annotation(attr.annotation)
                if attr.annotation
                else TypeAnnotation("Any")
            )

            if type_annotation.name == "Annotated":
                type_annotation_docstring = type_annotation.generic_args[1].name.strip("'")
                type_annotation = type_annotation.generic_args[0]
            else:
                type_annotation_docstring = None

            properties.append(
                Property(name=attr.target.id, type_annotation=type_annotation, docstring=type_annotation_docstring)
            )

    # Regular class with methods
    parsed_methods: List[Method] = []
    for method_node in methods:
        parsed_method = parse_method(method_node)

        if parsed_method:
            parsed_methods.append(parsed_method)

    return Class(name=name, docstring=docstring, is_enum=False, methods=parsed_methods, properties=properties)


def parse_function(func_node: ast.FunctionDef) -> Function:
    """Parse a standalone function."""
    parameters: List[Parameter] = []
    defaults = func_node.args.defaults
    # default_args = func_node.args.args[-len(defaults) :] if defaults else []

    for i, arg in enumerate(func_node.args.args):
        default_val = None
        if i >= len(func_node.args.args) - len(defaults):
            default_idx = i - (len(func_node.args.args) - len(defaults))
            if default_idx < len(defaults):
                default = defaults[default_idx]
                if isinstance(default, ast.Constant):
                    if isinstance(default.value, str):
                        default_val = f'"{default.value}"'
                    else:
                        default_val = repr(default.value)
                elif isinstance(default, ast.NameConstant):  # type: ignore
                    default_val = repr(default.value)
                elif isinstance(default, ast.Str):  # type: ignore Python 3.7 compatibility
                    default_val = f'"{default.s}"'  # type: ignore
                elif isinstance(default, ast.Name):
                    default_val = default.id
                else:
                    default_val = "..."

        param = parse_parameter(arg, default_val)
        parameters.append(param)

    return_type = None
    if func_node.returns:
        return_type = format_type_annotation(func_node.returns)

    docstring = extract_docstring(func_node)

    return Function(
        name=func_node.name,
        parameters=parameters,
        return_type=return_type,
        docstring=docstring,
    )


def parse_type_alias(assign_node: ast.Assign) -> Optional[TypeAlias]:
    """Parse a type alias assignment."""
    # Check if this is a simple assignment with a single target
    if len(assign_node.targets) != 1:
        return None

    target = assign_node.targets[0]
    if not isinstance(target, ast.Name):
        return None

    # Check if the value is a type annotation
    if not isinstance(assign_node.value, (ast.Name, ast.Attribute, ast.Subscript)):
        return None

    # Extract the name and type annotation
    name = target.id
    type_annotation = format_type_annotation(assign_node.value)

    # Look for docstring in the next node if it's a string literal
    docstring = None
    # Note: Type aliases typically don't have docstrings, but we could check
    # for comments or string literals that follow

    return TypeAlias(name=name, type_annotation=type_annotation, docstring=docstring)


def parse_module(file_path: Path) -> Module:
    """Parse a Python module file."""
    code = file_path.read_text()
    tree = ast.parse(code, filename=file_path.name)

    classes: List[Class] = []
    functions: List[Function] = []
    type_aliases: List[TypeAlias] = []

    for node in tree.body:
        if isinstance(node, ast.ClassDef):
            classes.append(parse_class(node))
        elif isinstance(node, ast.FunctionDef):
            functions.append(parse_function(node))
        elif isinstance(node, ast.Assign):
            # Check if this is a type alias
            type_alias = parse_type_alias(node)
            if type_alias:
                type_aliases.append(type_alias)

    return Module(
        file_path=file_path,
        classes=classes,
        functions=functions,
        type_aliases=type_aliases,
    )



from topk_sdk.data import binary_vector, f32_vector, u8_vector, i8_vector
from topk_sdk.query import field, fn, select

from . import ProjectContext
from .utils import dataset, doc_ids, is_sorted, doc_fields


def test_query_vector_distance(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(
            "title",
            summary_distance=fn.vector_distance("summary_embedding", [2.0] * 16),
        ).topk(field("summary_distance"), 3, True)
    )

    assert is_sorted(result, "summary_distance")
    assert doc_fields(result) == {"_id", "title", "summary_distance"}
    assert doc_ids(result) == {"1984", "pride", "mockingbird"}


def test_query_vector_distance_nullable(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(
            summary_distance=fn.vector_distance(
                "nullable_embedding", f32_vector([3.0] * 16)
            )
        ).topk(field("summary_distance"), 3, True)
    )

    assert is_sorted(result, "summary_distance")
    assert doc_ids(result) == {"1984", "mockingbird", "catcher"}


def test_query_vector_distance_u8_vector(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(
            summary_distance=fn.vector_distance("scalar_embedding", u8_vector([8] * 16))
        ).topk(field("summary_distance"), 3, True)
    )

    assert is_sorted(result, "summary_distance")
    assert doc_ids(result) == {"harry", "1984", "catcher"}


def test_query_vector_distance_i8_vector(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(
            summary_distance=fn.vector_distance(
                "scalar_i8_embedding", i8_vector([-10] * 16)
            )
        ).topk(field("summary_distance"), 3, True)
    )

    assert is_sorted(result, "summary_distance")
    assert doc_ids(result) == {"pride", "1984", "gatsby"}


def test_query_vector_distance_binary_vector(ctx: ProjectContext):
    collection = dataset.books.setup(ctx)

    result = ctx.client.collection(collection.name).query(
        select(
            summary_distance=fn.vector_distance(
                "binary_embedding", binary_vector([0, 1])
            )
        ).topk(field("summary_distance"), 2, True)
    )

    assert is_sorted(result, "summary_distance")
    assert doc_ids(result) == {"1984", "mockingbird"}
